import datetime

import pytz
from django.db import models


# TODO replace with something sensible

class Message(models.Model):
    text = models.CharField(max_length=255)
    uid = models.CharField(max_length=255)
    timestamp = models.BigIntegerField()
    is_response = models.BooleanField(default=False)  # true <=> sent by chatbot

    def get_time(self) -> str:
        return datetime.datetime.fromtimestamp(self.timestamp, tz=pytz.timezone('Europe/Prague')).strftime(
            '%Y-%m-%d %H:%M')

    def get_buttons(self) -> list:
        return list(Button.objects.all().filter(message_id__exact=self.id))

    def get_elements(self) -> list:
        return list(Element.objects.all().filter(message_id__exact=self.id))


class Button(models.Model):
    message = models.ForeignKey(Message, on_delete=models.CASCADE)
    text = models.CharField(max_length=255, null=True, blank=True)
    action = models.CharField(max_length=255, null=True, blank=True)  # json
    url = models.CharField(max_length=1024, null=True, blank=True)


class Element(models.Model):
    message = models.ForeignKey(Message, on_delete=models.CASCADE)
    title = models.CharField(max_length=255, null=True, blank=True)
    subtitle = models.CharField(max_length=255, null=True, blank=True)
    image_url = models.CharField(max_length=255, null=True, blank=True)



from django.apps import AppConfig

from golem.core.interfaces import all


class WebGui(AppConfig):
    name = 'webgui'
    prefix = 'web'
    verbose_name = "Chatbot web GUI"

    def ready(self):
        from .interface import WebGuiInterface
        all.register_chat_interface(WebGuiInterface)



import json
from datetime import datetime

import os

import random
import tensorflow as tf
import numpy as np
from django.conf import settings


class Model:
    def __init__(self, entity, base_dir):
        if not base_dir or not entity:
            raise ValueError('Base dir and entity name must be set')
        self.base_dir = base_dir
        self.entity = entity
        self.graph = tf.Graph()
        self.session = tf.Session(graph=self.graph)  # , config=tf.ConfigProto(log_device_placement=True))
        self.file_writer = tf.summary.FileWriter(os.path.join(base_dir, entity, "tensorboard", str(datetime.now())))
        self.loaded = False
        self.label_cnt = None

    def destroy(self):
        self.session.close()

    def make_weights(self, shape, name=None):
        return tf.Variable(tf.truncated_normal(shape, stddev=0.1), name=name)

    def make_biases(self, shape, name=None):
        return tf.Variable(tf.constant(0.1, shape=shape), name=name)

    def loss_fn(self):
        loss = tf.reduce_mean(
            tf.nn.softmax_cross_entropy_with_logits(
                labels=self.y, logits=self.y_
            )
        )
        return loss

    def get_last_component(self, lstm_output, length):
        batch_size = tf.shape(lstm_output)[0]
        max_length = tf.shape(lstm_output)[1]
        out_size = int(lstm_output.get_shape()[2])
        index = tf.range(0, batch_size) * max_length + (length - 1)
        flat = tf.reshape(lstm_output, [-1, out_size])
        relevant = tf.gather(flat, index)
        return relevant

    def init_network(self, feature_cnt, label_cnt, max_word_cnt, batch_size=1):
        memory_size = int(feature_cnt)

        self.x = tf.placeholder(tf.float32, shape=[None, max_word_cnt, feature_cnt], name='x')
        self.y = tf.placeholder(tf.float32, shape=[None, label_cnt], name='y')
        self.keep_prob = tf.placeholder_with_default(tf.cast(1.0, tf.float32), [], "keep_prob")
        self.lengths = tf.placeholder_with_default(tf.fill([batch_size], max_word_cnt), [batch_size], "lengths")

        with tf.name_scope("rnn"):

            fwd_lstm = tf.nn.rnn_cell.DropoutWrapper(
                tf.contrib.rnn.LSTMCell(memory_size),
                input_keep_prob=self.keep_prob,
                output_keep_prob=self.keep_prob,
                state_keep_prob=self.keep_prob
            )
            bwd_lstm = tf.nn.rnn_cell.DropoutWrapper(
                tf.contrib.rnn.LSTMCell(memory_size),
                input_keep_prob=self.keep_prob,
                output_keep_prob=self.keep_prob,
                state_keep_prob=self.keep_prob
            )

            rnn_1, state = tf.nn.bidirectional_dynamic_rnn(
                cell_fw=fwd_lstm,
                cell_bw=bwd_lstm,
                inputs=self.x,
                sequence_length=self.lengths,
                dtype=tf.float32
            )

            foo = self.get_last_component(rnn_1[0], self.lengths)
            bar = self.get_last_component(rnn_1[1], self.lengths)

        flat = tf.reshape(tf.stack([foo, bar], axis=2), [-1, memory_size * 2])

        dense_1 = tf.layers.dense(flat, label_cnt * 8, tf.nn.leaky_relu, name="dense_1")
        drop = tf.layers.dropout(dense_1, self.keep_prob)
        self.y_ = tf.layers.dense(drop, label_cnt, name="classifier")
        self.preds = tf.nn.softmax(self.y_)

    def train(self, batcher, num_iterations, keep_prob):

        with self.graph.as_default():

            feature_cnt = batcher.dim
            label_cnt = len(batcher.labels)
            max_word_cnt = batcher.max_words

            batch_size = settings.NLP_CONFIG.get("BATCH_SIZE", 32)
            initial_step = settings.NLP_CONFIG.get("INITIAL_STEP", 1e-3)
            use_tensorboard = settings.NLP_CONFIG.get("USE_TENSORBOARD", False)

            self.init_network(feature_cnt, label_cnt, max_word_cnt, batch_size=batch_size)
            loss = self.loss_fn()
            optimizer = tf.train.AdamOptimizer(initial_step).minimize(loss)
            accuracy = tf.reduce_mean(tf.metrics.accuracy(tf.argmax(self.preds, 1), tf.argmax(self.y, 1)))

            self.session.run(tf.global_variables_initializer())
            self.session.run(tf.local_variables_initializer())

            self.file_writer.add_graph(self.graph)
            tf.summary.scalar("loss", loss)
            tf.summary.scalar("accuracy", accuracy)
            tensorboard_op = tf.summary.merge_all() if use_tensorboard else None

            past_losses = [1.] * 25

            for i in range(num_iterations):

                # start_idx = random.randint(0, sample_cnt - batch_size - 1)
                # end_idx = start_idx + batch_size

                # _x_ = features[start_idx:end_idx]
                # _y_ = labels_onehot[start_idx:end_idx]
                # _junk_ = junk[start_idx:end_idx]

                # if i % 100 == 0:  TODO should this stay ?!
                #     total_junk_idx = random.randint(0, batch_size - 1)
                #     _x_[total_junk_idx] = [np.random.random(feature_cnt) for j in range(word_cnt)]
                #     _y_[total_junk_idx] = np.zeros(label_cnt)

                # trashify TODO already done in batcher
                # for j in range(len(_x_)):
                #     junks = _junk_[j]
                #     for junk_pos in junks:
                #         _x_[j][junk_pos] = np.random.random(feature_cnt)
                # print(_x_)

                # zipped = list(zip(_x_, _y_))
                # random.shuffle(zipped)
                # _x_, _y_ = zip(*zipped)

                _x_, _y_ = batcher.next_batch(batch_size)

                optimizer.run(feed_dict={self.x: _x_, self.y: _y_, self.keep_prob: keep_prob},
                              session=self.session)
                if use_tensorboard:
                    tensorboard_result = self.session.run(tensorboard_op, feed_dict={self.x: _x_, self.y: _y_})
                    self.file_writer.add_summary(tensorboard_result, i)

                xe = self.session.run(loss, feed_dict={self.x: _x_, self.y: _y_})

                print('Iteration ', i, '/', num_iterations,
                      'Xentropy:', xe,
                      'Accuracy:', self.session.run(accuracy, feed_dict={self.x: _x_, self.y: _y_}))

                past_losses.insert(0, xe)
                past_losses = past_losses[:25]
                if np.max(past_losses) < 0.05:
                    print("Early stopping, loss < 0.05")
                    break

            # save model
            saver = tf.train.Saver()
            saver.save(self.session, os.path.join(self.base_dir, self.entity, 'tf_session'))
            with open(os.path.join(self.base_dir, self.entity, 'tf_metadata.json'), 'w') as f:
                json.dump({'feature_cnt': feature_cnt, 'label_cnt': label_cnt, 'word_cnt': max_word_cnt}, f)

    def load(self):
        with open(os.path.join(self.base_dir, self.entity, 'tf_metadata.json')) as f:
            metadata = json.load(f)
        feature_cnt, label_cnt = metadata['feature_cnt'], metadata['label_cnt']
        word_cnt = metadata['word_cnt']
        print('features:', feature_cnt, 'labels:', label_cnt)
        print('Restoring from {}'.format(self.base_dir))
        self.init_network(feature_cnt, label_cnt, word_cnt)
        self.label_cnt = label_cnt
        saver = tf.train.Saver()
        saver.restore(self.session, os.path.join(self.base_dir, self.entity, 'tf_session'))

    def predict(self, features):
        with self.graph.as_default():
            if not self.loaded:
                self.load()
                self.loaded = True

            return self.session.run(self.preds, feed_dict={self.x: features})

    def test(self, x, y):
        with self.graph.as_default():
            if not self.loaded:
                self.load()
                self.loaded = True

            # accuracy = tf.metrics.accuracy(predictions=tf.argmax(self.y_, 1), labels=tf.argmax(self.y, 1))
            scores, predictions, losses = [], [], []
            output = []
            loss_fn = self.loss_fn()
            for _x_, _y_ in zip(x, y):
                labels = np.zeros(self.label_cnt)
                if _y_ != -1:
                    labels[_y_] = 1.0

                y_ = self.session.run(self.preds, feed_dict={self.x: _x_})
                loss = self.session.run(loss_fn, feed_dict={self.x: _x_, self.y: [labels]})
                arg = np.argmax(y_)
                predictions.append(arg)
                losses.append(loss)
                maxscore = np.max(y_)
                mean = np.mean(y_)
                if (arg == _y_ and maxscore > 0.8) or (maxscore < 0.8 and _y_ == -1):
                    print("Correct {}".format(maxscore))
                    scores.append(maxscore)
                    output.append((True, maxscore, arg))
                else:
                    print("Incorrect {}".format(maxscore))
                    scores.append(0.0)
                    output.append((False, maxscore, arg))
            correct = sorted(scores, reverse=True).index(0.0) if 0.0 in scores else len(scores)
            print("[{}] Accuracy score: {}% or {}/{}".format(self.entity, np.mean(scores) * 100,
                                                             correct, len(scores)))
            print("[{}] Mean Cross Entropy Loss: {}".format(self.entity, np.mean(losses)))
            return output, predictions



from django import forms


class MessageForm(forms.Form):
    message = forms.CharField(label='Message', max_length=1024,
                              widget=forms.TextInput(attrs={'placeholder': 'Type message here'}))



import json
import time

from golem.core.chat_session import ChatSession, Profile
from golem.core.persistence import get_elastic
from .responses.responses import MessageElement

class Logger:
    def __init__(self, session: ChatSession):
        self.uid = session.chat_id
        self.test_id = -1
        self.interface = session.interface
        self.enabled = bool(get_elastic())

    def log_user_message(self, type_, entities, accepted_time, state):
        #es = get_elastic()
        entities = entities.copy()
        text = None
        if '_message_text' in entities:
            text = entities['_message_text'][0]['value']
            del(entities['_message_text'])

        message = {
            'uid' : self.uid,
            'test_id' : self.test_id,
            'created': accepted_time,
            'is_user' : True,
            'text' : text,
            'state' : state,
            'type' : type_,
            'entities' : entities
        }
        self.log_message(message)

    def log_bot_message(self, response, state):
        #es = get_elastic()
        text = response.text if hasattr(response, 'text') else None
        if isinstance(response, str):
            text = response
            response = None
        elif not isinstance(response, MessageElement):
            return

        message = {
            'uid' : self.uid,
            'test_id' : self.test_id,
            'created': time.time(),
            'is_user' : False,
            'text' : text,
            'state' : state,
            'type' : type(response).__name__ if response else 'TextMessage',
            'response' : json.loads(json.dumps(response, default=lambda obj: obj.__dict__ if hasattr(obj,'__dict__') else str(obj)))
        }
        self.log_message(message)

    def log_error(self, exception, state):
        message = {
            'uid' : self.uid,
            'test_id' : self.test_id,
            'created': time.time(),
            'is_user' : False,
            'text' : str(exception),
            'state' : state,
            'type' : 'error'
        }
        self.log_message(message)

    def log_message(self, message):
        if not self.enabled:
            return
        es = get_elastic()
        if not es:
            return
        try:
            es.index(index="message-log", doc_type='message', body=message)
        except:
            print('Unable to log message to Elasticsearch.')

    def log_user(self, profile: Profile):
        if not self.enabled:
            return
        user = {
            'uid' : self.uid,
            'profile': profile.to_json()
        }
        es = get_elastic()
        if not es:
            return
        try:
            es.create(index="message-log", id=user['uid'], doc_type='user', body=user)
        except:
            print('Unable to log user profile to Elasticsearch.')



import json
import logging
import time

import random

from golem.core.chat_session import ChatSession
from golem.core.message_parser import parse_text_message
from golem.core.responses.buttons import PayloadButton
from golem.core.responses.quick_reply import LocationQuickReply
from golem.tasks import accept_user_message
from .models import Message, Button, Element


class WebGuiInterface:
    name = 'webgui'
    prefix = 'web'
    messages = []
    states = []

    @staticmethod
    def clear():
        WebGuiInterface.messages = []
        WebGuiInterface.states = []

    @staticmethod
    def load_profile(uid):
        return {'first_name': 'Tests', 'last_name': ''}

    @staticmethod
    def post_message(session, response):
        uid = session.meta.get("uid")
        WebGuiInterface.messages.append(response)
        message = Message()
        message.uid = uid
        message.timestamp = time.time()
        message.is_response = True
        try:
            message.text = response.text
        except AttributeError:
            pass
        message.save()
        if hasattr(response, 'buttons'):
            for btn in response.buttons:
                b = Button()
                b.message_id = message.id
                b.text = btn.title
                if hasattr(btn, 'url'):
                    b.action = 'link'
                    b.url = btn.url
                elif isinstance(btn, PayloadButton):
                    b.action = "postback"
                    b.url = btn.payload
                # todo parse stuff
                b.save()

        try:
            for q in response.quick_replies:
                b = Button()
                b.message_id = message.id
                if isinstance(q, LocationQuickReply):
                    b.text = 'My location\n(unsupported in web gui)'
                    b.action = 'null'  # TODO support sending location
                else:
                    b.text = q.title
                    b.action = 'reply'
                b.save()
        except AttributeError:
            pass
        try:
            for element in response.elements:
                e = Element()
                e.message_id = message.id
                e.title = element.title
                e.image_url = element.image_url
                e.subtitle = element.subtitle
                e.save()
                # todo buttons in elements etc.
                # todo quick replies
        except AttributeError:
            pass

    @staticmethod
    def send_settings(settings):
        pass

    @staticmethod
    def processing_start(session):
        pass

    @staticmethod
    def processing_end(session):
        pass

    @staticmethod
    def state_change(state):
        if not WebGuiInterface.states or WebGuiInterface.states[-1] != state:
            WebGuiInterface.states.append(state)

    @staticmethod
    def parse_message(user_message, num_tries=1):
        logging.info('[WEBGUI] @ parse_message')
        if user_message.get('text'):
            return parse_text_message(user_message.get('text'))
        elif user_message.get("payload"):
            data = user_message.get("payload")
            logging.info("Payload is: {}".format(data))
            if isinstance(data, dict):
                return {'entities': data, 'type': 'postback'}
            else:
                from golem.core.serialize import json_deserialize
                payload = json.loads(data, object_hook=json_deserialize)
                payload['_message_text'] = [{'value': None}]
                return {'entities': payload, 'type': 'postback'}

    @staticmethod
    def accept_request(msg: Message):
        uid = str(msg.uid)
        logging.info('[WEBGUI] Received message from {}'.format(uid))
        session = ChatSession(WebGuiInterface, uid, meta={"uid": uid})
        accept_user_message.delay(session.to_json(), {"text": msg.text})

    @staticmethod
    def accept_postback(msg: Message, data):
        uid = str(msg.uid)
        logging.info('[WEBGUI] Received postback from {}'.format(uid))
        session = ChatSession(WebGuiInterface, uid, meta={"uid": uid})
        accept_user_message.delay(session.to_json(), {"payload": data})

    @staticmethod
    def make_uid(username) -> str:
        uid = None
        tries = 0
        while (not uid or len(Message.objects.filter(uid__exact=uid)) != 0) or tries < 100:
            uid = '_'.join([str(username), str(random.randint(1000, 99999))])
            tries += 1
        # delete messages for old session with this uid, if there was one
        # FIXME invalidate the previous user's session!
        try:
            Message.objects.get(uid__exact=uid).delete()
        except Exception:
            pass  # first user, there is no such table yet
        return uid

    @staticmethod
    def destroy_uid(uid):
        Message.objects.filter(uid__exact=uid).delete()



import json
import pickle
import os

from django.conf import settings

from golem.nlp.keywords import keyword_search
from golem.nlp.nn.seq2seq import Seq2Seq

print('Will import GloVe and TF!')

import numpy as np
from golem.nlp import duckling, cleanup
from golem.nlp.nn.model import Model

from celery.app.log import get_logger

from golem.nlp import utils

logging = get_logger(__name__)

glove = utils.get_glove()
logging.debug('GloVe and TF imported')

models = {}
imputers = {}

NLP_DATA_DIR = utils.data_dir()
MAX_WORD_COUNT = settings.NLP_CONFIG.get("MAX_WORD_COUNT", 10)


def get_model(entity, entity_dir) -> Model:
    """Lazy loads a tf model."""
    if entity not in models:
        models[entity] = Model(entity, entity_dir)
        imputers[entity] = utils.get_imputation_rules(entity)
    return models[entity]


def word2vec(text, entity):
    """
    Converts entity to a vector for tensorflow.
    :param text     Text to be converted to BOW.
    :returns    numpy array of features
    """
    features = [np.zeros(glove.get_dimension()) for x in range(10)]
    tokens = cleanup.tokenize(text, stemming=False, language="cz")
    clean_text = cleanup.imputer(tokens, imputers.get(entity))
    print("Tokens:", clean_text)
    for idx, word in enumerate(clean_text):
        if idx > MAX_WORD_COUNT:
            logging.warning('Message too long!')  # FIXME allow longer messages
        vec = glove.get_vector(word)
        if vec is not None:
            features[idx] = vec
    return np.array([features])


def classify_trait(text, entity, threshold):
    """
    Classifies entity with a neural network.
    :returns:   Predicted label from tensorflow.
    """
    entity_dir = os.path.join(NLP_DATA_DIR, 'model', entity)
    pickle_data = pickle.load(open(os.path.join(entity_dir, 'pickle.json'), 'rb'))
    labels = pickle_data['labels']

    model = get_model(entity, entity_dir)
    bow = word2vec(text, entity)
    y_pred = model.predict(bow)[0]
    logging.debug(y_pred)
    print(y_pred)
    y_pred = [[i, score] for i, score in enumerate(y_pred) if score > threshold]
    y_pred.sort(key=lambda x: x[1], reverse=True)

    if len(y_pred) > 0:
        value = labels[y_pred[0][0]]
        prob = y_pred[0][1]
        print('> Predicted {}: {} Prob: {}%'.format(entity, value, prob * 100))
        print('All scores > threshold are:', y_pred)
        if value is not None and value != 'none':
            return [{'value': value}]
    return None


def classify(text: str, current_state=None):
    """
    Classifies all entity values of a text input.
    :param text
    :param current_state    State of the conversation without the init/accept suffix.
    :returns:   Dict containing found entities.
    """
    output = {}
    model_dir = os.path.join(NLP_DATA_DIR, 'model')
    dirs = [i for i in os.scandir(model_dir) if i.is_dir()]

    from golem.nlp import geneea
    try:
        text = geneea.get_correction(text)
        sentiment = geneea.get_sentiment(text)
    except:
        sentiment = 0.0

    for dir in dirs:
        # search in text for each known entity
        entity = dir.name
        entity_dir = os.path.join(model_dir, entity)

        with open(os.path.join(entity_dir, 'metadata.json'), 'r') as f:
            metadata = json.load(f)

        # if the entity is limited by allowed states, check if we're in one of them before continuing
        if current_state and 'allowed_states' in metadata:
            if current_state not in metadata['allowed_states']:
                continue

        if metadata['strategy'] == 'trait':
            pred = classify_trait(text, entity, metadata.get('threshold', 0.7))
            if pred:
                output[entity] = pred
        elif metadata['strategy'] == 'keywords':
            import unidecode
            with open(os.path.join(entity_dir, "trie.json"), 'r') as g:
                trie = json.load(g)
            should_stem = metadata.get('stemming', False)
            language = metadata.get('language', utils.get_default_language())
            pred = keyword_search(unidecode.unidecode(text), trie, should_stem, language)
            if pred:
                output[entity] = pred
        elif metadata['strategy'] == 'seq2seq':
            model = models.setdefault(entity, Seq2Seq(entity, entity_dir))
            pred = model.predict(text)
            if pred:
                output[entity] = [{'value': pred}]
        else:
            logging.warning('Unknown search strategy {} for entity {}, skipping!'
                            .format(metadata['strategy'], entity))

    if sentiment:
        output['sentiment'] = [{"value": sentiment}]
    output.update(duckling.get(text))
    logging.debug(output)
    return output


def test_all():
    test_dir = os.path.join(NLP_DATA_DIR, 'testing_data')
    entities = []
    for f in os.scandir(test_dir):
        name, ext = os.path.splitext(f.name)
        if f.is_file() and ext == '.json':
            entities.append((name, f))
    for entity, filename in entities:
        with open(filename) as f:
            examples = json.load(f).items()
            model_dir = os.path.join(NLP_DATA_DIR, 'model', entity)
            model = get_model(entity, model_dir)
            pickle_data = pickle.load(open(os.path.join(model_dir, 'pickle.json'), 'rb'))
            classes = pickle_data['labels']
            x = [word2vec(text, entity) for text, label in examples]
            for text, label in examples:
                print(text, label)
            y = [classes.index(label) if label is not None else -1 for text, label in examples]
            scores, predictions = model.test(x, y)
            for example, score_vec in zip(examples, scores):
                is_ok, score, pred = score_vec
                if not is_ok:
                    prediction = classes[pred] if score > 0.5 else None
                    print("Wrong label for '{}': '{}' expected: '{}'".format(example[0], prediction, example[1]))



from golem.core.parsing.entity_extractor import EntityExtractor


class GolemExtractor(EntityExtractor):

    def __init__(self):
        super().__init__()

    def extract_entities(self, text: str, max_retries=1):
        # TODO use a separate thread (pool) to remove TF memory overhead
        from golem.nlp.classify import classify
        return classify(text)



from golem.nlp.cleanup import tokenize


# TODO remove character accents

def prepare_keywords(data, should_stem=False, language='en') -> dict:
    """
    Prepares keywords for fast search.
    First words of each expression are processed into a trie.
    :param data:            A dict containing keyword data
    :param should_stem:     Boolean, whether to use stemmed keywords
                            for comparison. Default is false.
    :param language:        Stemming language. Default is 'en'.
    :return: trie ( dict )
    """
    # construct a trie from all chars of the keywords
    # may be thought of as finite state machine as well
    keywords = []
    root_state = {}
    for item in data:
        value = item.get('value')
        expressions = item.get('expressions', [])

        if not (value or expressions):
            raise ValueError("Keyword doesn't have any expressions")

        label = item.get('label', value)
        if not label:
            raise ValueError("Keyword doesn't have a label")

        if value:
            expressions.append(value)

        for expr in expressions:
            keywords.append((expr, label))

    for expression, label in keywords:
        words = tokenize(expression, should_stem, language=language)
        words = [str(word) for word in words]
        print(words)
        first_word = words[0]
        chars = list(first_word)
        state = root_state
        for idx, char in enumerate(chars):
            transitions = state.setdefault('transitions', {})
            state = transitions.setdefault(str(char), {})

        outputs = state.setdefault('outputs', [])
        outputs.append({
            "requires": words[1:],
            "label": label
        })

    return root_state


def keyword_search(text, trie, should_stem=False, language='en') -> list:
    """
    Searches for keyword matches in text.
    :param text:        Text that should be searched
    :param trie:        Trie containing the keyword representation
    :param should_stem: Boolean, whether to compare stemmed words.
                        Default is false.
    :param language:    Language for stemming. Default is 'en'.
    :return: A list of extracted entities, or an empty list.
    """
    extracted = []
    words = tokenize(text, should_stem, language)
    for idx, word in enumerate(words):
        state = trie
        chars = list(word)
        for char in chars:
            state = state.get('transitions', {}).get(char)
            if not state:
                break  # unknown word
        if state:
            outputs = state.get('outputs', [])
            outputs.sort(key=lambda x: len(x.get('requires', [])), reverse=True)
            following = set(words[idx + 1:])
            for output in outputs:
                required = set(output.get('requires', []))
                if required & following == required:
                    extracted.append({
                        "value": output.get('label')
                    })
    return extracted



import time
from datetime import datetime

from django.db.models import Max
from django.http import HttpResponse, HttpResponseBadRequest, JsonResponse
from django.shortcuts import render, redirect
from django.conf import settings

from .forms import MessageForm
from .interface import WebGuiInterface
from .models import Message


def webgui(request):
    if 'uid' in request.session:
        uid = request.session['uid']
        if request.method == 'POST':
            # message received via webgui
            if request.POST.get('message'):
                msg = Message()
                msg.uid = uid
                msg.text = request.POST.get('message')
                msg.timestamp = time.time()
                if msg.text is not None and len(msg.text) > 0:
                    # process message if text != null
                    msg.save()
                    if request.POST.get("postback"):
                        WebGuiInterface.accept_postback(msg, request.POST.get("postback"))
                    else:
                        WebGuiInterface.accept_request(msg)
            else:
                print("Error, message not set in POST")
                return HttpResponseBadRequest()
            return HttpResponse()

        messages = Message.objects.filter(uid=uid).order_by('timestamp')
        context = {
            'uid': uid, 'messages': messages,
            'form': MessageForm, 'timestamp': datetime.now().timestamp(),
            'user_img': settings.GOLEM_CONFIG.get('WEBGUI_USER_IMAGE', 'images/icon_user.png'),
            'bot_img': settings.GOLEM_CONFIG.get('WEBGUI_BOT_IMAGE', 'images/icon_robot.png')
        }
        return render(request, 'index.html', context)
    else:
        return render(request, 'welcome.html')


def do_login(request):
    if request.method == 'POST' and 'username' in request.POST:
        username = request.POST.get('username')
        uid = WebGuiInterface.make_uid(username)
        request.session['uid'] = uid
        request.session['username'] = username
        return HttpResponse()
    else:
        return HttpResponseBadRequest()


def do_logout(request):
    if 'uid' in request.session:
        WebGuiInterface.destroy_uid(request.session['uid'])
        del request.session['uid']
        del request.session['username']
    return redirect('webgui')


def get_last_change(request):
    if 'uid' in request.session:
        uid = request.session['uid']
        max_timestamp = Message.objects.filter(uid=uid).aggregate(Max('timestamp'))
        return JsonResponse(max_timestamp)
    return HttpResponseBadRequest()



from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = 'Tests the NLP on a sentence'

    # def add_arguments(self, parser):
    #     parser.add_argument('text', nargs=1, type=str)

    def handle(self, *args, **options):
        from golem.nlp import classify
        classify.test_all()
